import gzip
import json
import os
import sys
import tempfile
from pathlib import Path
from unittest import mock

import responses  # type: ignore

from launchable.utils.http_client import get_base_url
from launchable.utils.session import read_session, write_build
from tests.cli_test_case import CliTestCase


class CTestTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/ctest/').resolve()

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_multiple_files(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(), self.organization, self.workspace),
            json={
                "testPaths": [
                    [{'type': 'testcase', 'name': 'FooTest.Bar'}],
                    [{'type': 'testcase', 'name': 'FooTest.Foo'}],
                    [{'type': 'testcase', 'name': '*/ParameterizedTest.Bar/*'}],
                ],
                "rest": [
                    [{'name': 'FooTest.Baz'}],
                ],
                "subsettingId": 456,
                "summary": {
                    "subset": {"candidates": 4, "duration": 4, "rate": 90},
                    "rest": {"candidate": 1, "duration": 0, "rate": 0}
                },
                "isBrainless": False
            },
            status=200)
        with tempfile.TemporaryDirectory() as tempdir:
            # Use a non-existing dir to check it creates a dir.
            output_dir = os.path.join(tempdir, 'subdir')

            # emulate launchable record build
            write_build(self.build_name)

            result = self.cli('subset', '--target', '10%', 'ctest',
                              '--output-regex-files',
                              '--output-regex-files-dir=' + output_dir,
                              '--output-regex-files-size=32',
                              str(self.test_files_dir.joinpath("ctest_list.json")))
            self.assertEqual(result.exit_code, 0)

            subset_files = []
            rest_files = []
            for file in os.listdir(output_dir):
                with open(os.path.join(output_dir, file), 'r') as f:
                    if file.startswith('subset'):
                        subset_files.append(f.read().strip())
                    else:
                        rest_files.append(f.read().strip())
            subset_files.sort()
            rest_files.sort()
            if sys.version_info[:2] >= (3, 7):
                self.assertEqual(subset_files, [
                    '^FooTest\\.Bar$|^FooTest\\.Foo$',
                    '^\\*/ParameterizedTest\\.Bar/\\*$'
                ])
            else:
                # There was a change in re.escape behavior from Python 3.7. See
                # https://docs.python.org/3/library/re.html#re.escape
                self.assertEqual(subset_files, [
                    '^FooTest\\.Bar$|^FooTest\\.Foo$',
                    '^\\*\\/ParameterizedTest\\.Bar\\/\\*$'
                ])
            self.assertEqual(rest_files, ['^FooTest\\.Baz$'])

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_without_session(self):
        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli('subset', '--target', '10%', 'ctest', str(self.test_files_dir.joinpath("ctest_list.json")))
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())
        expected = self.load_json_from_file(self.test_files_dir.joinpath('subset_result.json'))
        self.assert_json_orderless_equal(payload, expected)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_test(self):
        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli('record', 'tests', 'ctest', str(self.test_files_dir) + "/Testing/**/Test.xml")
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(read_session(self.build_name), self.session)

        payload = json.loads(gzip.decompress(responses.calls[2].request.body).decode())
        expected = self.load_json_from_file(self.test_files_dir.joinpath('record_test_result.json'))

        for c in payload['events']:
            del c['created_at']
        self.assert_json_orderless_equal(payload, expected)



# TODO: add cli-specific custom exceptions

class ParseSessionException(Exception):
    def __init__(
        self,
        session: str,
        message: str = "Wrong session format; session format is like 'builds/<build name>/test_sessions/<test session id>'.",
    ):
        self.session = session
        self.message = "{message}: {session}".format(message=message, session=self.session)
        super().__init__(self.message)


class InvalidJUnitXMLException(Exception):
    def __init__(
        self,
        filename: str,
        message: str = "Invalid JUnit XML file format",
    ):
        self.filename = filename
        self.message = "{message}: {filename}".format(message=message, filename=self.filename)
        super().__init__(self.message)



"""Path name matching with extended GLOBs.

Primarily developed to interface with Maven, which supports "**", "*", and "?" as the special characters
"""
import re
from typing import Pattern


def is_path_separator(c: str):
    return c == '/' or c == '\\'


def compile(glob: str) -> Pattern:
    """Compiles a glob pattern like foo/**/*.txt into a """
    # fnmatch.fnmatch is close but it doesn't deal with paths well, including
    # '**'

    p = ""
    i = 0
    n = len(glob)
    while i < n:
        c = glob[i]
        i += 1

        if c == '*':
            if i < n and glob[i] == '*':
                i += 1
                if i < n and is_path_separator(glob[i]):
                    # '**/' matches any sub-directories or none
                    i += 1
                    p += "(.+[\\/])?"
                else:
                    # '**' used like **Test.java. Is this even legal?
                    p += ".*"
            else:
                p += "[^\\/]*"
        elif c == '?':
            p += "[^\\/]"
        elif is_path_separator(c):
            p += "[\\/]"
        else:
            p += re.escape(c)

    return re.compile(p)



import gzip
import json
import os
import tempfile
from pathlib import Path
from unittest import mock

import responses  # type: ignore

from launchable.utils.http_client import get_base_url
from tests.cli_test_case import CliTestCase
from tests.helper import ignore_warnings


class MinitestTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/minitest/').resolve()
    result_file_path = test_files_dir.joinpath('record_test_result.json')

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_test_minitest(self):
        result = self.cli('record', 'tests', '--session', self.session, 'minitest', str(self.test_files_dir) + "/")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())

        expected = self.load_json_from_file(self.result_file_path)
        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_test_minitest_chunked(self):
        result = self.cli('record', 'tests', '--session', self.session,
                          '--post-chunk', 5, 'minitest', str(self.test_files_dir) + "/")
        self.assertEqual(result.exit_code, 0)

        payload1 = json.loads(gzip.decompress(responses.calls[1].request.body).decode())
        expected1 = self.load_json_from_file(self.test_files_dir.joinpath('record_test_result_chunk1.json'))

        payload2 = json.loads(gzip.decompress(responses.calls[2].request.body).decode())
        expected2 = self.load_json_from_file(self.test_files_dir.joinpath('record_test_result_chunk2.json'))

        # concat 1st and 2nd request for unstable order
        self.assert_json_orderless_equal(expected1['events'] + expected2['events'],
                                         payload1['events'] + payload2['events'])

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    @ignore_warnings
    def test_subset(self):
        test_path = Path("test", "example_test.rb")
        responses.replace(responses.POST,
                          "{}/intake/organizations/{}/workspaces/{}/subset".format(get_base_url(),
                                                                                   self.organization,
                                                                                   self.workspace),
                          json={'testPaths': [[{'name': str(test_path)}]],
                                'rest': [],
                                'subsettingId': 123,
                                'summary': {
                                    'subset': {
                                        'duration': 10,
                                        'candidates': 1,
                                        'rate': 100},
                                    'rest': {
                                        'duration': 0,
                                        'candidates': 0,
                                        'rate': 0}},
                                "isBrainless": False,
                                },
                          status=200)

        result = self.cli('subset', '--target', '20%', '--session', self.session, '--base', str(self.test_files_dir),
                          'minitest', str(self.test_files_dir) + "/test/**/*.rb")

        self.assertEqual(result.exit_code, 0)
        output = Path(self.test_files_dir, "test", "example_test.rb")
        self.assertIn(str(output), result.output)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_with_invalid_path(self):
        result = self.cli('subset', '--target', '20%', '--session', self.session, '--base', str(self.test_files_dir),
                          'minitest', str(self.test_files_dir) + "/dummy")

        self.assertEqual(result.exit_code, 0)
        self.assertTrue("Error: no tests found matching the path." in result.output)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    @ignore_warnings
    def test_subset_split(self):
        test_path = Path("test", "example_test.rb")
        responses.replace(responses.POST,
                          "{}/intake/organizations/{}/workspaces/{}/subset".format(get_base_url(),
                                                                                   self.organization,
                                                                                   self.workspace),
                          json={'testPaths': [[{'name': str(test_path)}]],
                                'rest': [],
                                'subsettingId': 123,
                                'summary': {'subset': {'duration': 10,
                                                       'candidates': 1,
                                                       'rate': 100},
                                            'rest': {'duration': 0,
                                                     'candidates': 0,
                                                     'rate': 0}},
                                "isBrainless": False,
                                },
                          status=200)

        result = self.cli('subset', '--target', '20%', '--session', self.session, '--base',
                          str(self.test_files_dir), '--split', 'minitest', str(self.test_files_dir) + "/test/**/*.rb")

        self.assertEqual(result.exit_code, 0)
        self.assertIn('subset/123', result.output)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    @ignore_warnings
    def test_split_subset(self):
        test_path = Path("test", "example_test.rb")
        responses.replace(responses.POST,
                          "{}/intake/organizations/{}/workspaces/{}/subset/456/slice".format(get_base_url(),
                                                                                             self.organization,
                                                                                             self.workspace),
                          json={'testPaths': [[{'name': str(test_path)}]],
                                'rest': [],
                                'subsettingId': 123},
                          status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli('split-subset', '--subset-id', 'subset/456', '--base', str(self.test_files_dir),
                          '--bin', '2/2', '--rest', rest.name, 'minitest')

        self.assertEqual(result.exit_code, 0)
        output = Path(self.test_files_dir, "test", "example_test.rb")
        self.assertEqual(str(output), result.output.rstrip("\n"))
        self.assertEqual(rest.read().decode().rstrip("\n"), str(output))
        rest.close()
        os.unlink(rest.name)



import os
import pathlib
import subprocess
import sys
import urllib
from typing import Dict, List, Optional, Tuple

# Path component is a node in a tree.
# It's the equivalent of a short file/directory name in a file system.
# In our abstraction, it's represented as arbitrary bag of attributes
TestPathComponent = Dict[str, str]

# TestPath is a full path to a node in a tree from the root
# It's the equivalent of an absolute file name in a file system
TestPath = List[TestPathComponent]


def parse_test_path(tp_str: str) -> TestPath:
    """Parse a string representation of TestPath."""
    if tp_str == '':
        return []
    ret = []  # type: TestPath
    for component_str in tp_str.split('#'):
        if component_str == '&':
            # Technically, this should be mapped to {None:None}. But because the
            # TestPath definition is now Dict[str, str], not Dict[Optional[str],
            # Optinal[str]], we cannot add it. Fixing this definition needs to
            # fix callers not to assume they are always str. In practice, this
            # is a rare case. Do not appent {None: None} now...
            # ret.append({None: None})
            continue
        first = True
        component = {}
        for kv in component_str.split('&'):
            if first:
                first = False
                if kv:
                    (component['type'], component['name']) = _parse_kv(kv)
            else:
                (k, v) = _parse_kv(kv)
                component[k] = v
        ret.append(component)
    return ret


def _parse_kv(kv: str) -> Tuple[str, str]:
    kvs = kv.split('=')
    if len(kvs) != 2:
        raise ValueError('Malformed TestPath component: ' + kv)
    return (_decode_str(kvs[0]), _decode_str(kvs[1]))


def unparse_test_path(tp: TestPath) -> str:
    """Create a string representation of TestPath."""
    ret = []
    for component in tp:
        s = ''
        pairs = []
        if component.get('type', None) and component.get('name', None):
            s += _encode_str(component['type']) + '=' + _encode_str(component['name'])
            for k, v in component.items():
                if k not in ('type', 'name'):
                    pairs.append((k, v))
        else:
            for k, v in component.items():
                if not k or not v:
                    continue
                pairs.append((k, v))
            if len(pairs) == 0:
                s = '&'
        pairs = sorted(pairs, key=lambda p: p[0])
        for (k, v) in pairs:
            s += '&'
            s += _encode_str(k) + '=' + _encode_str(v)
        ret.append(s)
    return '#'.join(ret)


def _decode_str(s: str) -> str:
    return urllib.parse.unquote(s)


def _encode_str(s: str) -> str:
    return s.replace('%', '%25').replace('=', '%3D').replace('#', '%23').replace('&', '%26')


def _relative_to(p: pathlib.Path, base: str) -> pathlib.Path:
    if sys.version_info[0:2] >= (3, 6):
        return p.resolve(strict=False).relative_to(base)
    else:
        try:
            resolved = p.resolve()
        except BaseException:
            resolved = p
        return resolved.relative_to(base)


class FilePathNormalizer:
    """Normalize file paths based on the Git repository root

    Some test runners output absolute file paths. This is not preferrable when
    making statistical data on tests as the absolute paths can vary per machine
    or per run. FilePathNormalizer guesses the relative paths based on the Git
    repository root.
    """

    def __init__(self, base_path: Optional[str] = None, no_base_path_inference: bool = False):
        self._base_path = base_path
        self._no_base_path_inference = no_base_path_inference
        self._inferred_base_path = None  # type: Optional[str]

    def relativize(self, p: str) -> str:
        return str(self._relativize(pathlib.Path(os.path.normpath(p))))

    def _relativize(self, p: pathlib.Path) -> pathlib.Path:
        if not p.is_absolute():
            return p

        if self._base_path:
            return _relative_to(p, self._base_path)

        if self._no_base_path_inference:
            return p

        if not self._inferred_base_path:
            self._inferred_base_path = self._auto_infer_base_path(p)

        if self._inferred_base_path:
            return _relative_to(p, self._inferred_base_path)

        return p

    def _auto_infer_base_path(self, p: pathlib.Path) -> Optional[str]:
        p = p.parent
        while p != p.root and not p.exists():
            p = p.parent
        try:
            toplevel = subprocess.check_output(
                ['git', 'rev-parse', '--show-superproject-working-tree'],
                cwd=str(p),
                stderr=subprocess.DEVNULL,
                universal_newlines=True).strip()
            if toplevel:
                return toplevel
            return subprocess.check_output(
                ['git', 'rev-parse', '--show-toplevel'],
                cwd=str(p),
                stderr=subprocess.DEVNULL,
                universal_newlines=True).strip()
        except subprocess.CalledProcessError:
            # Cannot infer the Git repo. Continue with the abs path...
            return None



import gzip
import json
import os
import tempfile
from pathlib import Path
from unittest import mock

import responses  # type: ignore

from launchable.utils.http_client import get_base_url
from tests.cli_test_case import CliTestCase
from tests.helper import ignore_warnings


class NUnitTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/nunit/').resolve()

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset(self):
        responses.replace(
            responses.POST, "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(), self.organization, self.workspace), json={
                'testPaths': [
                    [
                        {
                            "type": "Assembly",
                            "name": "calc.dll",
                        },
                        {
                            "type": "TestSuite",
                            "name": "ParameterizedTests",
                        },
                        {
                            "type": "TestFixture",
                            "name": "MyTests",
                        },
                        {
                            "type": "ParameterizedMethod",
                            "name": "DivideTest",
                        },
                        {
                            "type": "TestCase",
                            "name": "DivideTest(12,3)",
                        },
                    ],
                    [
                        {
                            "type": "Assembly",
                            "name": "calc.dll",
                        },
                        {
                            "type": "TestSuite",
                            "name": "calc",
                        },
                        {
                            "type": "TestFixture",
                            "name": "Tests1",
                        },
                        {
                            "type": "TestCase",
                            "name": "Test1",
                        },
                    ],
                ],
                'rest': [],
                'subsettingId': 123,
                'summary': {
                    'subset': {
                        'duration': 15,
                        'candidates': 2,
                        'rate': 100,
                    },
                    'rest': {
                        'duration': 0,
                        'candidates': 0,
                        'rate': 0,
                    },
                },
            }, status=200)

        result = self.cli('subset', '--target', '10%', '--session', self.session, 'nunit',
                          str(self.test_files_dir) + "/list.xml")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[0].request.body).decode())

        expected = self.load_json_from_file(self.test_files_dir.joinpath('subset_result.json'))

        self.assert_json_orderless_equal(expected, payload)

        output = 'ParameterizedTests.MyTests.DivideTest(12,3)\ncalc.Tests1.Test1'
        self.assertIn(output, result.output)

    @ignore_warnings
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_split_subset(self):
        responses.replace(
            responses.POST, "{}/intake/organizations/{}/workspaces/{}/subset/456/slice".format(
                get_base_url(), self.organization, self.workspace), json={
                'testPaths': [
                    [
                        {
                            "type": "Assembly", "name": "calc.dll",
                        },
                        {
                            "type": "TestSuite", "name": "ParameterizedTests",
                        },
                        {
                            "type": "TestFixture", "name": "MyTests",
                        },
                        {
                            "type": "ParameterizedMethod", "name": "DivideTest",
                        },
                        {
                            "type": "TestCase", "name": "DivideTest(12,3)",
                        },
                    ],
                ],
                'rest': [
                    [
                        {
                            "type": "Assembly", "name": "calc.dll",
                        },
                        {
                            "type": "TestSuite", "name": "calc",
                        },
                        {
                            "type": "TestFixture", "name": "Tests1",
                        },
                        {
                            "type": "TestCase", "name": "Test1",
                        },
                    ],
                ],
                'subsettingId': 456,
                'summary': {
                    'subset': {
                        'duration': 8, 'candidates': 1, 'rate': 50,
                    },
                    'rest': {
                        'duration': 7, 'candidates': 1, 'rate': 50,
                    },
                },
            },
            status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli('split-subset', '--subset-id', 'subset/456',
                          '--bin', '1/2', '--rest', rest.name, 'nunit')

        self.assertEqual(result.exit_code, 0)

        self.assertIn('ParameterizedTests.MyTests.DivideTest(12,3)', result.output)

        self.assertEqual(rest.read().decode(), 'calc.Tests1.Test1')
        rest.close()
        os.unlink(rest.name)

    @responses.activate
    @mock.patch.dict(os.environ,
                     {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_test_on_linux(self):
        result = self.cli('record', 'tests', '--session', self.session,
                          'nunit', str(self.test_files_dir) + "/output-linux.xml")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())

        expected = self.load_json_from_file(self.test_files_dir.joinpath("record_test_result.json"))

        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_test_on_windows(self):
        result = self.cli('record', 'tests', '--session', self.session,
                          'nunit', str(self.test_files_dir) + "/output-windows.xml")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())

        expected = self.load_json_from_file(self.test_files_dir.joinpath("record_test_result.json"))

        self.assert_json_orderless_equal(expected, payload)



import gzip
import json
import os
from pathlib import Path
from unittest import mock

import responses  # type: ignore

from launchable.test_runners.pytest import _parse_pytest_nodeid
from tests.cli_test_case import CliTestCase


class PytestTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/pytest/').resolve()
    result_file_path = test_files_dir.joinpath('record_test_result.json')
    subset_input = '''tests/funcs3_test.py::test_func4
tests/funcs3_test.py::test_func5
tests/test_funcs1.py::test_func1
tests/test_funcs1.py::test_func2
tests/test_funcs2.py::test_func3
tests/test_funcs2.py::test_func4
tests/test_mod.py::TestClass::test__can_print_aaa
tests/fooo/func4_test.py::test_func6
tests/fooo/filenameonly_test.py

9 tests collected in 0.02s
'''

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset(self):
        result = self.cli('subset', '--target', '10%', '--session',
                          self.session, 'pytest', input=self.subset_input)
        self.assertEqual(result.exit_code, 0)
        payload = json.loads(gzip.decompress(responses.calls[0].request.body).decode())
        expected = self.load_json_from_file(self.test_files_dir.joinpath('subset_result.json'))
        for test_path in expected["testPaths"]:
            test_path[0]['name'] = os.path.normpath(test_path[0]['name'])
        self.assertEqual(expected, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_test_pytest(self):
        result = self.cli('record', 'tests', '--session', self.session,
                          'pytest', str(self.test_files_dir.joinpath("report.xml")))

        self.assertEqual(result.exit_code, 0)
        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())
        expected = self.load_json_from_file(self.result_file_path)
        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_test_with_json_option(self):
        result = self.cli('record', 'tests', '--session', self.session,
                          'pytest', '--json', str(self.test_files_dir.joinpath("report.json")))

        self.assertEqual(result.exit_code, 0)
        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())
        expected = self.load_json_from_file(self.result_file_path)

        for e in payload["events"]:
            e.pop("created_at", "")
        for e in expected["events"]:
            e.pop("created_at", "")

        self.assert_json_orderless_equal(expected, payload)

    def setUp(self):
        super().setUp()
        self.current_dir = os.getcwd()
        os.chdir(str(self.test_files_dir))

    def tearDown(self):
        os.chdir(self.current_dir)
        super().tearDown()

    def test_parse_pytest_nodeid(self):

        self.assertEqual(_parse_pytest_nodeid("tests/test_mod.py::TestClass::test__can_print_aaa"), [
            {"type": "file", "name": os.path.normpath("tests/test_mod.py")},
            {"type": "class", "name": "tests.test_mod.TestClass"},
            {"type": "testcase", "name": "test__can_print_aaa"},
        ])

        self.assertEqual(_parse_pytest_nodeid("tests/fooo/func4_test.py::test_func6"), [
            {"type": "file", "name": os.path.normpath(
                "tests/fooo/func4_test.py")},
            {"type": "class", "name": "tests.fooo.func4_test"},
            {"type": "testcase", "name": "test_func6"},
        ])



import os
import platform
from unittest import TestCase, mock

from launchable.utils.http_client import LaunchableClient
from launchable.version import __version__


class LaunchableClientTest(TestCase):
    @mock.patch.dict(
        os.environ,
        {"LAUNCHABLE_ORGANIZATION": "launchableinc", "LAUNCHABLE_WORKSPACE": "test"},
        clear=True,
    )
    def test_header(self):
        cli = LaunchableClient("/test")
        self.assertEqual(cli._headers(True), {
            'Content-Encoding': 'gzip',
            'Content-Type': 'application/json',
            "User-Agent": "Launchable/{} (Python {}, {})".format(
                __version__,
                platform.python_version(),
                platform.platform(),
            ),
        })

        self.assertEqual(cli._headers(False), {
            'Content-Type': 'application/json',
            "User-Agent": "Launchable/{} (Python {}, {})".format(
                __version__,
                platform.python_version(),
                platform.platform(),
            ),
        })

        cli = LaunchableClient("/test", test_runner="dummy")
        self.assertEqual(cli._headers(False), {
            'Content-Type': 'application/json',
            "User-Agent": "Launchable/{} (Python {}, {}) TestRunner/{}".format(
                __version__,
                platform.python_version(),
                platform.platform(),
                "dummy",
            ),
        })



import os
import sys
from http import HTTPStatus
from typing import Dict, List, Mapping, Optional

import click

from ...utils.ci_provider import CIProvider
from ...utils.click import KeyValueType
from ...utils.env_keys import REPORT_ERROR_KEY
from ...utils.flavor import normalize_flavors
from ...utils.http_client import LaunchableClient
from ...utils.session import write_session

LAUNCHABLE_SESSION_DIR_KEY = 'LAUNCHABLE_SESSION_DIR'

JENKINS_URL_KEY = 'JENKINS_URL'
JENKINS_BUILD_URL_KEY = 'BUILD_URL'
GITHUB_ACTIONS_KEY = 'GITHUB_ACTIONS'
GITHUB_ACTIONS_SERVER_URL_KEY = 'GITHUB_SERVER_URL'
GITHUB_ACTIONS_REPOSITORY_KEY = 'GITHUB_REPOSITORY'
GITHUB_ACTIONS_RUN_ID_KEY = 'GITHUB_RUN_ID'
CIRCLECI_KEY = 'CIRCLECI'
CIRCLECI_BUILD_URL_KEY = 'CIRCLE_BUILD_URL'


@click.command()
@click.option(
    '--build',
    'build_name',
    help='build name',
    required=True,
    type=str,
    metavar='BUILD_NAME'
)
@click.option(
    '--save-file/--no-save-file',
    'save_session_file',
    help='save session to file',
    default=True,
    metavar='SESSION_FILE'
)
@click.option(
    "--flavor",
    "flavor",
    help='flavors',
    cls=KeyValueType,
    multiple=True,
)
@click.option(
    "--observation",
    "is_observation",
    help="enable observation mode",
    is_flag=True,
)
@click.pass_context
def session(
    ctx: click.core.Context,
    build_name: str,
    save_session_file: bool,
    print_session: bool = True,
    flavor: List[str] = [],
    is_observation: bool = False,
):
    """
    print_session is for barckward compatibility.
    If you run this `record session` standalone,
    the command should print the session ID because v1.1 users expect the beheivior.
    That is why the flag is default True.
    If you run this command from the other command such as `subset` and `record tests`,
    you should set print_session = False because users don't expect to print session ID to the subset output.
    """

    flavor_dict = {}
    for f in normalize_flavors(flavor):
        flavor_dict[f[0]] = f[1]

    link = _capture_link(os.environ)
    payload = {
        "flavors": flavor_dict,
        "isObservation": is_observation,
    }
    if link:
        payload["link"] = link

    client = LaunchableClient(dry_run=ctx.obj.dry_run)
    try:
        sub_path = "builds/{}/test_sessions".format(build_name)
        res = client.request("post", sub_path, payload=payload)

        if res.status_code == HTTPStatus.NOT_FOUND:
            click.echo(
                click.style(
                    "Build {} was not found. "
                    "Make sure to run `launchable record build --name {}` before you run this command.".format(
                        build_name,
                        build_name),
                    'yellow'),
                err=True,
            )
            sys.exit(1)

        res.raise_for_status()
        session_id = res.json()['id']

        if save_session_file:
            write_session(build_name, "{}/{}".format(sub_path, session_id))
        if print_session:
            # what we print here gets captured and passed to `--session` in
            # later commands
            click.echo("{}/{}".format(sub_path, session_id))

    except Exception as e:
        if os.getenv(REPORT_ERROR_KEY):
            raise e
        else:
            click.echo(e, err=True)


def _capture_link(env: Mapping[str, str]) -> Optional[Dict[str, str]]:
    if env.get(JENKINS_URL_KEY):
        return {"provider": CIProvider.JENKINS.value, "url": env.get(JENKINS_BUILD_URL_KEY, "")}
    elif env.get(GITHUB_ACTIONS_KEY):
        return {"provider": CIProvider.GITHUB_ACTIONS.value, "url": "{}/{}/actions/runs/{}".format(
            env.get(GITHUB_ACTIONS_SERVER_URL_KEY),
            env.get(GITHUB_ACTIONS_REPOSITORY_KEY),
            env.get(GITHUB_ACTIONS_RUN_ID_KEY),
        ),
        }
    elif env.get(CIRCLECI_KEY):
        return {"provider": CIProvider.CIRCLECI.value, "url": env.get(CIRCLECI_BUILD_URL_KEY, "")}
    else:
        return None



import gzip
import json
import os
import sys
from pathlib import Path
from unittest import mock

import responses  # type: ignore

from launchable.commands.record.tests import INVALID_TIMESTAMP, parse_launchable_timeformat
from launchable.utils.session import write_build, write_session
from tests.cli_test_case import CliTestCase


class TestsTest(CliTestCase):
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_with_group_name(self):
        # emulate launchable record build & session
        write_session(self.build_name, self.session_id)

        report_files_dir = Path(__file__).parent.joinpath(
            '../../data/maven/').resolve()

        result = self.cli('record', 'tests', '--session',
                          self.session, '--group', 'hoge', 'maven', str(
                              report_files_dir) + "**/reports/")

        self.assertEqual(result.exit_code, 0)
        # get request body
        # responses.calls[0]: GET: build information
        # responses.calls[1]: POST: record tests
        request = json.loads(gzip.decompress(
            responses.calls[1].request.body).decode())

        self.assertCountEqual(request.get("group", []), "hoge")

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_filename_in_error_message(self):
        # emulate launchable record build
        write_build(self.build_name)

        normal_xml = str(Path(__file__).parent.joinpath('../../data/broken_xml/normal.xml').resolve())
        broken_xml = str(Path(__file__).parent.joinpath('../../data/broken_xml/broken.xml').resolve())
        result = self.cli('record', 'tests', '--build', self.build_name, 'file', normal_xml, broken_xml)

        def remove_backslash(input: str) -> str:
            # Hack for Windowns. They containts double escaped backslash such
            # as \\\\
            if sys.platform == "win32":
                return input.replace("\\", "")
            else:
                return input

        # making sure the offending file path name is being printed.
        self.assertIn(remove_backslash(broken_xml), remove_backslash(result.output))

        # normal.xml
        self.assertIn('open_class_user_test.rb', gzip.decompress(responses.calls[2].request.body).decode())

    def test_parse_launchable_timeformat(self):
        t1 = "2021-04-01T09:35:47.934+00:00"  # 1617269747.934
        t2 = "2021-05-24T18:29:04.285+00:00"  # 1621880944.285
        t3 = "2021-05-32T26:29:04.285+00:00"  # invalid time format

        parse_launchable_time1 = parse_launchable_timeformat(t1)
        parse_launchable_time2 = parse_launchable_timeformat(t2)

        self.assertEqual(parse_launchable_time1.timestamp(), 1617269747.934)
        self.assertEqual(parse_launchable_time2.timestamp(), 1621880944.285)

        self.assertEqual(INVALID_TIMESTAMP, parse_launchable_timeformat(t3))



from typing import List, Sequence, Tuple


def normalize_flavors(flavors_raw: List[str]) -> List[Tuple[str, str]]:
    """Normalize flavor list, because the type of the flavor list specified in the command line flags differs depending
    on the version of the click.

    TODO: handle extraction of flavor tuple to dict in better way for >=click8.0 that returns tuple of tuples as tuple
    of str
    E.G.
        <click8.0:
            `launchable record session --build aaa --flavor os=ubuntu --flavor python=3.5`
            is parsed as build=aaa, flavor=(("os", "ubuntu"), ("python", "3.5"))
        >=click8.0:
            `launchable record session --build aaa --flavor os=ubuntu --flavor python=3.8`
            is parsed as build=aaa, flavor=("('os', 'ubuntu')", "('python', '3.8')")
    """
    flavors = []
    for f in flavors_raw:
        if isinstance(f, str):
            k, v = f.replace("(", "").replace(")", "").replace("'", "").split(",")
            flavors.append((k.strip(), v.strip()))
        elif isinstance(f, Sequence):
            flavors.append((f[0], f[1]))

    return flavors



import gzip
import json
import os
import tempfile
from pathlib import Path
from unittest import mock

import responses  # type: ignore

from launchable.utils.http_client import get_base_url
from launchable.utils.session import write_build
from tests.cli_test_case import CliTestCase
from tests.helper import ignore_warnings


class GradleTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/gradle/').resolve()
    result_file_path = test_files_dir.joinpath('recursion/expected.json')

    @ignore_warnings
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_without_session(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(),
                self.organization,
                self.workspace),
            json={
                "testPaths": [
                    [{'name': 'com.launchableinc.rocket_car_gradle.App2Test'}],
                    [{'name': 'com.launchableinc.rocket_car_gradle.AppTest'}],
                    [{'name': 'com.launchableinc.rocket_car_gradle.sub.App3Test'}],
                    [{'name': 'com.launchableinc.rocket_car_gradle.utils.UtilsTest'}]
                ],
                "rest": [],
                "subsettingId": 456,
                "summary": {
                    "subset": {"candidates": 4, "duration": 4, "rate": 100},
                    "rest": {"candidate": 0, "duration": 0, "rate": 0},
                },
                "isBrainless": False},
            status=200)

        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli(
            'subset', '--target', '10%', 'gradle',
            str(self.test_files_dir.joinpath('java/app/src/test').resolve()))
        # TODO: we need to assert on the request payload to make sure it found
        # test list all right
        self.assertEqual(result.exit_code, 0)
        output = "--tests com.launchableinc.rocket_car_gradle.App2Test " \
                 "--tests com.launchableinc.rocket_car_gradle.AppTest " \
                 "--tests com.launchableinc.rocket_car_gradle.sub.App3Test " \
                 "--tests com.launchableinc.rocket_car_gradle.utils.UtilsTest"
        self.assertIn(output, result.output.rstrip('\n'))

    @ignore_warnings
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_rest(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(),
                self.organization,
                self.workspace,
            ),
            json={
                "testPaths": [
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.App2Test'}],
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.AppTest'}],
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.utils.UtilsTest'}],
                ],
                "rest": [[{'name': 'com.launchableinc.rocket_car_gradle.sub.App3Test'}]],
                "subsettingId": 456,
                "summary": {
                    "subset": {"candidates": 3, "duration": 3, "rate": 75},
                    "rest": {"candidate": 1, "duration": 1, "rate": 25},
                },
                "isBrainless": False,
            },
            status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)

        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli(
            'subset', '--target', '10%', '--rest', rest.name, 'gradle',
            str(self.test_files_dir.joinpath('java/app/src/test/java').resolve()))

        self.assertEqual(result.exit_code, 0)
        self.assertIn("--tests com.launchableinc.rocket_car_gradle.App2Test "
                      "--tests com.launchableinc.rocket_car_gradle.AppTest "
                      "--tests com.launchableinc.rocket_car_gradle.utils.UtilsTest",
                      result.output.rstrip('\n'))
        self.assertEqual(rest.read().decode(), '--tests com.launchableinc.rocket_car_gradle.sub.App3Test')
        rest.close()
        os.unlink(rest.name)

    @ignore_warnings
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_zero_input_subsetting(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(),
                self.organization,
                self.workspace),
            json={
                "testPaths": [
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.AppTest'}],
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.utils.UtilsTest'}],
                ],
                "rest": [
                    [{'name': 'com.launchableinc.rocket_car_gradle.sub.App2Test'}],
                    [{'name': 'com.launchableinc.rocket_car_gradle.sub.App3Test'}],
                ],
                "subsettingId": 456,
                "summary": {
                    "subset": {"candidates": 3, "duration": 3, "rate": 75},
                    "rest": {"candidate": 1, "duration": 1, "rate": 25}
                },
                "isBrainless": False,
            },
            status=200)

        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli(
            'subset',
            '--target',
            '10%',
            '--get-tests-from-previous-sessions',
            '--output-exclusion-rules',
            'gradle',
            mix_stderr=False)

        if result.exit_code != 0:
            self.assertEqual(
                result.exit_code,
                0,
                "Exit code is not 0. The output is\n" + result.output + "\n" + result.stderr)
        subset_arg = result.output.rstrip('\n')
        self.assertEqual(
            subset_arg,
            "-PexcludeTests=com/launchableinc/rocket_car_gradle/sub/App2Test.class,com/launchableinc/rocket_car_gradle/"
            "sub/App3Test.class")

    @ignore_warnings
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_zero_input_subsetting_observation(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(),
                self.organization,
                self.workspace),
            json={
                "testPaths": [
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.AppTest'}],
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.utils.UtilsTest'}],
                ],
                "rest": [
                    [{'name': 'com.launchableinc.rocket_car_gradle.sub.App2Test'}],
                    [{'name': 'com.launchableinc.rocket_car_gradle.sub.App3Test'}],
                ],
                "subsettingId": 456,
                "summary": {
                    "subset": {"candidates": 2, "duration": 3, "rate": 75},
                    "rest": {"candidate": 2, "duration": 1, "rate": 25}
                },
                "isBrainless": False,
                "isObservation": True,
            },
            status=200)

        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli(
            'subset',
            '--target',
            '10%',
            '--get-tests-from-previous-sessions',
            '--output-exclusion-rules',
            'gradle',
            mix_stderr=False)

        if result.exit_code != 0:
            self.assertEqual(
                result.exit_code,
                0,
                "Exit code is not 0. The output is\n" + result.output + "\n" + result.stderr)
        subset_arg = result.output.rstrip('\n')
        self.assertEqual(subset_arg, "-PexcludeTests=")

    @ignore_warnings
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_zero_input_subsetting_source_root(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(),
                self.organization,
                self.workspace),
            json={
                "testPaths": [
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.AppTest'}],
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.utils.UtilsTest'}],
                ],
                "rest": [
                    [{'name': 'com.launchableinc.rocket_car_gradle.sub.App2Test'}],
                    [{'name': 'com.launchableinc.rocket_car_gradle.sub.App3Test'}],
                ],
                "subsettingId": 456,
                "summary": {
                    "subset": {"candidates": 2, "duration": 3, "rate": 75},
                    "rest": {"candidate": 2, "duration": 1, "rate": 25}
                },
                "isBrainless": False,
                "isObservation": True,
            },
            status=200)

        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli(
            'subset',
            '--target',
            '10%',
            '--get-tests-from-previous-sessions',
            '--output-exclusion-rules',
            'gradle',
            str(self.test_files_dir.joinpath('java/app/src/test').resolve()),
            mix_stderr=False)

        if result.exit_code != 0:
            self.assertEqual(
                result.exit_code,
                0,
                "Exit code is not 0. The output is\n" + result.output + "\n" + result.stderr)

        body = gzip.decompress(responses.calls[1].request.body).decode('utf8')
        self.assertNotIn("java.com.launchableinc.rocket_car_gradle.App2Test", body)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_split(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(),
                self.organization,
                self.workspace),
            json={
                "testPaths": [
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.App2Test'}],
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.AppTest'}],
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.utils.UtilsTest'}],
                ],
                "rest": [[{'name': 'com.launchableinc.rocket_car_gradle.sub.App3Test'}]],
                "subsettingId": 123,
                "summary": {
                    "subset": {"candidates": 3, "duration": 3, "rate": 75},
                    "rest": {"candidate": 1, "duration": 1, "rate": 25}
                },
                "isBrainless": False,
            },
            status=200)

        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli(
            'subset',
            '--target',
            '10%',
            '--split',
            'gradle',
            str(self.test_files_dir.joinpath('java/app/src/test/java').resolve()))

        self.assertEqual(result.exit_code, 0)
        self.assertIn("subset/123", result.output.rstrip('\n'))

    @ignore_warnings
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_split_subset(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset/456/slice".format(
                get_base_url(),
                self.organization,
                self.workspace),
            json={
                'testPaths': [
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.App2Test'}],
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.AppTest'}],
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.utils.UtilsTest'}],
                ],
                "rest": [[{'name': 'com.launchableinc.rocket_car_gradle.sub.App3Test'}]],
            },
            status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli(
            'split-subset',
            '--subset-id',
            'subset/456',
            '--bin',
            '1/2',
            '--rest',
            rest.name,
            'gradle')

        self.assertEqual(result.exit_code, 0)
        self.assertIn(
            "--tests com.launchableinc.rocket_car_gradle.App2Test "
            "--tests com.launchableinc.rocket_car_gradle.AppTest "
            "--tests com.launchableinc.rocket_car_gradle.utils.UtilsTest",
            result.output.rstrip('\n'))
        self.assertEqual(rest.read().decode(), '--tests com.launchableinc.rocket_car_gradle.sub.App3Test')
        rest.close()
        os.unlink(rest.name)

    @ignore_warnings
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_split_subset_with_same_bin(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset/456/slice".format(
                get_base_url(),
                self.organization,
                self.workspace,
            ),
            json={
                'testPaths': [
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.App2Test'}],
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.AppTest'}],
                    [{'type': 'class',
                      'name': 'com.launchableinc.rocket_car_gradle.utils.UtilsTest'}],
                ],
                "rest": [],
            },
            status=200,
        )

        same_bin_file = tempfile.NamedTemporaryFile(delete=False)
        same_bin_file.write(
            b'com.launchableinc.rocket_car_gradle.App2Test\n'
            b'com.launchableinc.rocket_car_gradle.AppTest\n'
            b'com.launchableinc.rocket_car_gradle.utils.UtilsTest')
        result = self.cli(
            'split-subset',
            '--subset-id',
            'subset/456',
            '--bin',
            '1/2',
            "--same-bin",
            same_bin_file.name,
            'gradle',
        )

        self.assertEqual(result.exit_code, 0)
        self.assertIn(
            "--tests com.launchableinc.rocket_car_gradle.App2Test "
            "--tests com.launchableinc.rocket_car_gradle.AppTest "
            "--tests com.launchableinc.rocket_car_gradle.utils.UtilsTest",
            result.output.rstrip('\n'),
        )
        same_bin_file.close()
        os.unlink(same_bin_file.name)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_test_gradle(self):
        result = self.cli('record', 'tests', '--session', self.session,
                          'gradle', str(self.test_files_dir) + "/**/reports")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())

        expected = self.load_json_from_file(self.result_file_path)
        self.assert_json_orderless_equal(expected, payload)



import os
import subprocess
import sys
from typing import List, Optional
from urllib.parse import urlparse

import click

from ...utils.commit_ingester import upload_commits
from ...utils.env_keys import REPORT_ERROR_KEY
from ...utils.git_log_parser import parse_git_log
from ...utils.http_client import get_base_url
from ...utils.java import cygpath, get_java_command
from ...utils.logger import LOG_LEVEL_AUDIT, Logger

jar_file_path = os.path.normpath(os.path.join(os.path.dirname(__file__), "../../jar/exe_deploy.jar"))


@click.command()
@click.option(
    '--source',
    help="repository path",
    default=os.getcwd(),
    type=click.Path(exists=True, file_okay=False),
)
@click.option(
    '--executable',
    help="[Obsolete] it was to specify how to perform commit collection but has been removed",
    type=click.Choice(['jar', 'docker']),
    default='jar',
    hidden=True)
@click.option(
    '--max-days',
    help="the maximum number of days to collect commits retroactively",
    default=30)
@click.option(
    '--scrub-pii',
    is_flag=True,
    help='[Deprecated] Scrub emails and names',
    hidden=True)
@click.option(
    '--import-git-log-output',
    help="import from the git-log output",
    type=click.Path(exists=True, dir_okay=False,
                    resolve_path=True, allow_dash=True),
)
@click.pass_context
def commit(ctx, source: str, executable: bool, max_days: int, scrub_pii: bool, import_git_log_output: str):
    if executable == 'docker':
        sys.exit("--executable docker is no longer supported")

    if import_git_log_output:
        _import_git_log(import_git_log_output, ctx.obj.dry_run)
        return

    try:
        exec_jar(os.path.abspath(source), max_days, ctx.obj.dry_run)
    except Exception as e:
        if os.getenv(REPORT_ERROR_KEY):
            raise e
        else:
            click.echo(click.style(
                "Can't get commit history from `{}`. Do you run command root of git-controlled directory? "
                "If not, please set a directory use by --source option."
                .format(os.path.abspath(source)),
                fg='yellow'),
                err=True)
            print(e)


def exec_jar(source: str, max_days: int, dry_run: bool):
    java = get_java_command()

    if not java:
        sys.exit("You need to install Java")

    base_url = get_base_url()

    # using subprocess.check_out with shell=False and a list of command to prevent vulnerability
    # https://knowledge-base.secureflag.com/vulnerabilities/code_injection/os_command_injection_python.html
    command = [java]
    command.extend(_build_proxy_option(os.getenv("HTTPS_PROXY")))
    command.extend([
        "-jar",
        cygpath(jar_file_path),
        "ingest:commit",
        "-endpoint",
        "{}/intake/".format(base_url),
        "-max-days",
        str(max_days),
        "-scrub-pii",
    ])

    if Logger().logger.isEnabledFor(LOG_LEVEL_AUDIT):
        command.append("-audit")
    if dry_run:
        command.append("-dry-run")
    command.append(cygpath(source))

    subprocess.run(
        command,
        check=True,
        shell=False,
    )


def _import_git_log(output_file: str, dry_run: bool):
    try:
        with click.open_file(output_file) as fp:
            commits = parse_git_log(fp)
        upload_commits(commits, dry_run)
    except Exception as e:
        if os.getenv(REPORT_ERROR_KEY):
            raise e
        else:
            click.echo(
                click.style("Failed to import the git-log output", fg='yellow'),
                err=True)
            print(e)


def _build_proxy_option(https_proxy: Optional[str]) -> List[str]:
    if not https_proxy:
        return []

    if not (https_proxy.startswith("https://") or https_proxy.startswith("http://")):
        https_proxy = "https://" + https_proxy
    proxy_url = urlparse(https_proxy)

    options = []
    if proxy_url.hostname:
        options.append("-Dhttps.proxyHost={}".format(proxy_url.hostname))
    if proxy_url.port:
        options.append("-Dhttps.proxyPort={}".format(proxy_url.port))
    return options



import os
import platform
import re
import subprocess
from typing import List

import click

from ..utils.authentication import get_org_workspace
from ..utils.click import emoji
from ..utils.http_client import LaunchableClient
from ..utils.java import get_java_command
from ..version import __version__ as version


def compare_version(a: List[int], b: List[int]):
    """Compare two version numbers represented as int arrays"""

    def pick(a, i):
        return a[i] if i < len(a) else 0

    for i in range(max(len(a), len(b))):
        d = pick(a, i) - pick(b, i)
        if d != 0:
            return d  # if they are different, we have the result
    return 0  # identical


def compare_java_version(output: str) -> int:
    """Check if the Java version meets what we need. returns >=0 if we meet the requirement"""
    for l in output.splitlines():
        if l.find("java version") != -1:
            # l is like: java version "1.8.0_144"
            m = re.search('"([^"]+)"', l)
            if m:
                tokens = m.group(1).split(".")
                if len(tokens) >= 2:
                    versions = [int(x) for x in tokens[0:2]]
                    required = [1, 8]
                    return compare_version(versions, required)
    # couldn't determine, so err on the safe side
    return 0


def check_java_version(javacmd: str) -> int:
    """Check if the Java version meets what we need. returns >=0 if we meet the requirement"""
    v = subprocess.run([javacmd, "-version"], check=True, stderr=subprocess.PIPE, universal_newlines=True)
    return compare_java_version(v.stderr)


@click.command(name="verify")
def verify():
    # In this command, regardless of REPORT_ERROR_KEY, always report an unexpected error with full stack trace
    # to assist troubleshooting. `click.UsageError` is handled by the invoking
    # Click gracefully.

    org, workspace = get_org_workspace()
    client = LaunchableClient()
    java = get_java_command()

    # Print the system information first so that we can get them even if there's
    # an issue.

    click.echo("Organization: " + repr(org))
    click.echo("Workspace: " + repr(workspace))
    click.echo("Proxy: " + repr(os.getenv("HTTPS_PROXY")))
    click.echo("Platform: " + repr(platform.platform()))
    click.echo("Python version: " + repr(platform.python_version()))
    click.echo("Java command: " + repr(java))
    click.echo("launchable version: " + repr(version))

    if org is None or workspace is None:
        raise click.UsageError(
            click.style(
                "Could not identify Launchable organization/workspace. "
                "Please confirm if you set LAUNCHABLE_TOKEN or LAUNCHABLE_ORGANIZATION and LAUNCHABLE_WORKSPACE "
                "environment variables",
                fg="red"))

    res = client.request("get", "verification")
    if res.status_code == 401:
        raise click.UsageError(click.style("Authentication failed. Most likely the value for the LAUNCHABLE_TOKEN "
                                           "environment variable is invalid.", fg="red"))
    res.raise_for_status()

    if java is None:
        raise click.UsageError(click.style(
            "Java is not installed. Install Java version 8 or newer to use the Launchable CLI.", fg="red"))

    # Level 2 check: versions. This is more fragile than just reporting the number, so we move
    # this out here

    if compare_version([int(x) for x in platform.python_version().split('.')], [3, 5]) < 0:
        raise click.UsageError(click.style("Python 3.5 or later is required", fg="red"))

    if check_java_version(java) < 0:
        raise click.UsageError(click.style("Java 8 or later is required", fg="red"))

    click.echo(click.style("Your CLI configuration is successfully verified" + emoji(" \U0001f389"), fg="green"))



import gzip
import json
import os
import tempfile
from unittest import mock

import responses  # type: ignore

from launchable.utils.http_client import get_base_url
from launchable.utils.session import write_build
from tests.cli_test_case import CliTestCase
from tests.helper import ignore_warnings


class RawTest(CliTestCase):
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(get_base_url(), self.organization, self.workspace),
            json={
                "testPaths": [
                    [{'type': 'testcase', 'name': 'FooTest.Bar'}],
                    [{'type': 'testcase', 'name': 'FooTest.Foo'}],
                ],
                "testRunner": "raw",
                "rest": [
                    [{'name': 'FooTest.Baz'}],
                ],
                "subsettingId": 456,
                "summary": {
                    "subset": {"candidates": 4, "duration": 4, "rate": 90},
                    "rest": {"candidate": 1, "duration": 0, "rate": 0}
                },
                "isBrainless": False
            },
            status=200)
        with tempfile.TemporaryDirectory() as tempdir:
            test_path_file = os.path.join(tempdir, 'tests.txt')
            with open(test_path_file, 'w') as f:
                f.write('\n'.join([
                    'testcase=FooTest.Bar',
                    'testcase=FooTest.Foo',
                    '# This is a comment',
                    'testcase=FooTest.Baz',
                ]) + '\n')

            # emulate launchable record build
            write_build(self.build_name)

            result = self.cli('subset', '--target', '10%',
                              'raw', test_path_file, mix_stderr=False)
            self.assertEqual(result.exit_code, 0)

            # Check request body
            payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())
            self.assert_json_orderless_equal(payload, {
                'testPaths': [
                    [{'type': 'testcase', 'name': 'FooTest.Bar'}],
                    [{'type': 'testcase', 'name': 'FooTest.Foo'}],
                    [{'type': 'testcase', 'name': 'FooTest.Baz'}]
                ],
                'testRunner': 'raw',
                'session': {'id': str(self.session_id)},
                "goal": {"type": "subset-by-percentage", "percentage": 0.1},
                "ignoreNewTests": False,
                "getTestsFromPreviousSessions": False,
            })
            # Check split output
            self.assertEqual(result.stdout, '\n'.join([
                'testcase=FooTest.Bar',
                'testcase=FooTest.Foo',
            ]) + '\n')

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_get_tests_from_previous_sessions(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(), self.organization, self.workspace),
            json={
                "testPaths": [
                    [{'type': 'testcase', 'name': 'FooTest.Bar'}],
                    [{'type': 'testcase', 'name': 'FooTest.Foo'}],
                ],
                "testRunner": "raw",
                "rest": [
                    [{'type': 'testcase', 'name': 'FooTest.Baz'}],
                ],
                "subsettingId": 456,
                "summary": {
                    "subset": {"candidates": 4, "duration": 4, "rate": 90},
                    "rest": {"candidate": 1, "duration": 0, "rate": 0}
                },
                "isBrainless": False
            },
            status=200)

        # emulate launchable record build
        write_build(self.build_name)

        # Don't use with for Windows environment
        rest = tempfile.NamedTemporaryFile(mode="+w", encoding="utf-8", delete=False)
        result = self.cli(
            'subset',
            '--target',
            '10%',
            '--get-tests-from-previous-sessions',
            "--rest", rest.name,
            'raw',
            mix_stderr=False)
        self.assertEqual(result.exit_code, 0)

        # Check request body
        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())
        self.assert_json_orderless_equal(payload, {
            'testPaths': [],
            'testRunner': 'raw',
            'session': {'id': str(self.session_id)},
            "goal": {"type": "subset-by-percentage", "percentage": 0.1},
            "ignoreNewTests": False,
            "getTestsFromPreviousSessions": True,
        })

        # Check outputs
        self.assertEqual(result.stdout, "testcase=FooTest.Bar\ntestcase=FooTest.Foo" + "\n")
        self.assertEqual(rest.read(), "testcase=FooTest.Baz")

        rest.close()
        os.unlink(rest.name)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_tests(self):
        with tempfile.TemporaryDirectory() as tempdir:
            test_path_file = os.path.join(tempdir, 'tests.json')
            with open(test_path_file, 'w') as f:
                f.write('\n'.join([
                    '{',
                    '  "testCases": [',
                    '     {',
                    '       "testPath": "file=a.py#class=classA",',
                    '       "duration": 42,',
                    '       "status": "TEST_PASSED",',
                    '       "stdout": "This is stdout",',
                    '       "stderr": "This is stderr",',
                    '       "createdAt": "2021-10-05T12:34:00"',
                    '     }',
                    '  ]',
                    '}',
                ]) + '\n')

            # emulate launchable record build
            write_build(self.build_name)

            result = self.cli('record', 'tests', 'raw', test_path_file, mix_stderr=False)
            self.assertEqual(result.exit_code, 0)

            # Check request body
            payload = json.loads(gzip.decompress(responses.calls[2].request.body).decode())
            self.assert_json_orderless_equal(payload, {
                'events': [
                    {
                        'testPath': [
                            {'type': 'file', 'name': 'a.py'},
                            {'type': 'class', 'name': 'classA'},
                        ],
                        'duration': 42,
                        'status': 1,
                        'stdout': 'This is stdout',
                        'stderr': 'This is stderr',
                        'created_at': '2021-10-05T12:34:00',
                        'data': None,
                        'type': 'case',
                    },
                ],
                "testRunner": "raw",
                "group": ""
            })

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_tests_junit_xml(self):
        with tempfile.TemporaryDirectory() as tempdir:
            test_path_file = os.path.join(tempdir, 'tests.xml')
            with open(test_path_file, 'w') as f:
                f.write('\n'.join([
                    '<?xml version="1.0" encoding="UTF-8"?>',
                    '<testsuites name="test_suite_name1" tests="1" failures="0" errors="0" time="10.123">',
                    '  <testsuite name="test_suite_name2" errors="0" failures="0" skipped="0" timestamp="2021-10-05T12:34:00" time="10.123" tests="1">',   # noqa: E501
                    '    <testcase classname="test_class_name" name="test_case_name" time="10.123">',
                    '    </testcase>',
                    '  </testsuite>',
                    '</testsuites>',
                ]) + '\n')

            # emulate launchable record build
            write_build(self.build_name)

            result = self.cli('record', 'tests', 'raw', test_path_file, mix_stderr=False)
            if result.exit_code != 0:
                self.assertEqual(
                    result.exit_code,
                    0,
                    "Exit code is not 0. The output is\n" + result.output + "\n" + result.stderr)

            # Check request body
            payload = json.loads(gzip.decompress(responses.calls[2].request.body).decode())
            self.assert_json_orderless_equal(payload, {
                'events': [
                    {
                        'testPath': [
                            {'type': 'class', 'name': 'test_class_name'},
                            {'type': 'testcase', 'name': 'test_case_name'},
                        ],
                        'duration': 10.123,
                        'status': 1,
                        'stdout': '',
                        'stderr': '',
                        'created_at': '2021-10-05T12:34:00',
                        'data': None,
                        'type': 'case',
                    },
                ],
                "testRunner": "raw",
                "group": ""
            })

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    @ignore_warnings
    def test_split_subset(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset/456/slice".format(
                get_base_url(), self.organization, self.workspace),
            json={
                "testPaths": [
                    [{'type': 'testcase', 'name': 'FooTest.Bar'}],
                    [{'type': 'testcase', 'name': 'FooTest.Foo'}],
                ],
                "rest": [[{'type': 'testcase', 'name': 'FooTest.Baz'}]],
            },
            status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli(
            'split-subset',
            '--subset-id',
            'subset/456',
            '--bin',
            '1/2',
            '--rest',
            rest.name,
            'raw')

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(
            result.stdout,
            '\n'.join([
                'testcase=FooTest.Bar',
                'testcase=FooTest.Foo\n',
            ]))
        self.assertEqual(
            rest.read().decode(),
            'testcase=FooTest.Baz',
        )
        rest.close()
        os.unlink(rest.name)

    @responses.activate
    @mock.patch.dict(os.environ,
                     {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_split_subset_with_same_bin(self):
        # This test must raise error.
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset/456/slice".format(
                get_base_url(), self.organization, self.workspace),
            json={
                "testPaths": [
                    [{'type': 'testcase', 'name': 'FooTest.Bar'}],
                    [{'type': 'testcase', 'name': 'FooTest.Foo'}],
                ],
                "rest": [],
            },
            status=200)

        same_bin_file = tempfile.NamedTemporaryFile(delete=False)
        same_bin_file.write(
            b'FooTest.Bar\n'
            b'FooTest.Foo')
        result = self.cli(
            'split-subset',
            '--subset-id',
            'subset/456',
            '--bin',
            '1/2',
            "--same-bin",
            same_bin_file.name,
            'raw')
        self.assertTrue("--same-bin option is supported only for gradle test and go-test." in result.stdout)
        same_bin_file.close()
        os.unlink(same_bin_file.name)



import glob
import os
import re
from typing import Dict, List

import click
from junitparser import TestCase, TestSuite  # type: ignore

from ..testpath import TestPath
from ..utils.logger import Logger
from . import launchable


@launchable.subset
def subset(client):
    logger = Logger()

    # NOTE: This should be using package name + test function name to specify
    # which tests to run. However, the initial integration is created so that we
    # only specify the test function names. This can result in matching some
    # extra tests in multiple packages. However, in order to keep the initial
    # way of the integration, we cannot change this. Try to do the best.
    test_cases = []
    for line in client.stdin():
        if ' ' not in line:
            test_cases.append(line.strip('\n'))
        else:
            parts = re.split('\\s+', line)
            if len(parts) >= 2:
                package = parts[1].split('/')[-1]
                for test_case in test_cases:
                    client.test_path([{'type': 'class', 'name': package}, {
                                     'type': 'testcase', 'name': test_case}])
            else:
                logger.warning("Cannot extract the package from the input. This may result in missing some tests.")
            test_cases = []
    client.formatter = lambda x: "^{}$".format(x[1]['name'])
    client.separator = '|'
    client.run()


@click.argument('source_roots', required=True, nargs=-1)
@launchable.record.tests
def record_tests(client, source_roots):
    for root in source_roots:
        match = False
        for t in glob.iglob(root, recursive=True):
            match = True
            if os.path.isdir(t):
                client.scan(t, "*.xml")
            else:
                client.report(t)

        if not match:
            click.echo("No matches found: {}".format(root), err=True)
            return

    default_path_builder = client.path_builder

    def path_builder(case: TestCase, suite: TestSuite, report_file: str) -> TestPath:
        tp = default_path_builder(case, suite, report_file)
        for tpc in tp:
            if 'type' in tpc and 'name' in tpc and tpc['type'] == 'class':
                # go-junit-report v2 reports full package name. go-junit-report
                # v1 reports only the last component of the package name. In
                # order to make this backward compatible, we align this to v1.
                tpc['name'] = tpc['name'].split('/')[-1]
        return tp

    client.path_builder = path_builder
    client.run()


def format_same_bin(s: str) -> List[Dict[str, str]]:
    t = s.split(".")
    return [{"type": "class", "name": t[0]},
            {"type": "testcase", "name": t[1]}]


split_subset = launchable.CommonSplitSubsetImpls(
    __name__,
    formatter=lambda x: "^{}$".format(x[1]['name']),
    seperator='|',
    same_bin_formatter=format_same_bin,
).split_subset()



import glob
import os
from typing import Dict, List

import click

from launchable.commands.record.case_event import CaseEvent
from launchable.testpath import TestPath
from launchable.utils.sax import Element, SaxParser, TagMatcher

from . import launchable

# common code between 'subset' & 'record tests' to build up test path from
# nested <test-suite>s


def build_path(e: Element):
    pp = []  # type: TestPath
    if e.parent:
        pp = e.parent.tags.get('path') or []    # type: ignore
    if e.name == "test-suite":
        # <test-suite>s form a nested tree structure so capture those in path
        pp = pp + [{'type': e.attrs['type'], 'name': e.attrs['name']}]
    if e.name == "test-case":
        pp = pp + [{'type': 'TestCase', 'name': e.attrs['name']}]

    if len(pp) > 0:
        def split_filepath(path: str) -> List[str]:
            # Supports Linux and Windows
            if '/' in path:
                return path.split('/')
            else:
                return path.split('\\')

        # "Assembly" type containts full path at a cutomer's environment
        # remove file path prefix in Assembly
        e.tags['path'] = [
            {**path, 'name': split_filepath(path['name'])[-1]}
            if path['type'] == 'Assembly'
            else path
            for path in pp
        ]


@click.argument('report_xmls', type=click.Path(exists=True), required=True, nargs=-1)
@launchable.subset
def subset(client, report_xmls):
    """
    Parse an XML file produced from NUnit --explore option to list up all the viable test cases
    """

    def on_element(e: Element):
        build_path(e)
        if e.name == "test-case":
            client.test_path(_replace_fixture_to_suite(e.tags['path']))

    for report_xml in report_xmls:
        SaxParser([], on_element).parse(report_xml)

    # join all the names except when the type is ParameterizedMethod, because in that case test cases have
    # the name of the test method in it and ends up creating duplicates
    client.formatter = lambda x: '.'.join([c['name'] for c in x if c['type'] not in ['ParameterizedMethod', 'Assembly']])
    client.run()


split_subset = launchable.CommonSplitSubsetImpls(__name__, formatter=lambda x: '.'.join(
    [c['name'] for c in x if c['type'] not in ['ParameterizedMethod', 'Assembly']])).split_subset()


@click.argument('report_xml', required=True, nargs=-1)
@launchable.record.tests
def record_tests(client, report_xml):
    # TODO: copied from ctest -- promote this pattern
    for root in report_xml:
        match = False
        for t in glob.iglob(root, recursive=True):
            match = True
            if os.path.isdir(t):
                client.scan(t, "*.xml")
            else:
                client.report(t)
        if not match:
            click.echo("No matches found: {}".format(root), err=True)

    def parse_func(report: str):
        events = []

        # parse <test-case> element into CaseEvent
        def on_element(e: Element):
            build_path(e)
            if e.name == "test-case":
                events.append(CaseEvent.create(
                    _replace_fixture_to_suite(e.tags['path']),  # type: ignore
                    float(e.attrs['duration']),
                    CaseEvent.TEST_PASSED if e.attrs['result'] == 'Passed' else CaseEvent.TEST_FAILED,
                    timestamp=str(e.tags['startTime'])))  # timestamp is already iso-8601 formatted

        # the 'start-time' attribute is normally on <test-case> but apparently not always,
        # so we try to use the nearest ancestor as an approximate
        SaxParser([TagMatcher.parse("*/@start-time={startTime}")], on_element).parse(report)

        # return the obtained events as a generator
        return (x for x in events)

    client.parse_func = parse_func
    client.run()


"""
    Nunit produces different XML structure report between without --explore option and without it.
    So we replace TestFixture to TestSuite to avid this difference problem.
"""


def _replace_fixture_to_suite(paths) -> List[Dict[str, str]]:
    for path in paths:
        if path["type"] == "TestFixture":
            path["type"] = "TestSuite"

    return paths



import os
from typing import List, Optional

import click

from ..utils.click import FRACTION, FractionType
from ..utils.env_keys import REPORT_ERROR_KEY
from ..utils.http_client import LaunchableClient
from .test_path_writer import TestPathWriter


@click.group(help="Split subsetting tests")
@click.option(
    '--subset-id',
    'subset_id',
    help='subset id',
    type=str,
    required=True,
)
@click.option(
    '--bin',
    'bin_target',
    help='bin',
    type=FRACTION,
    required=True
)
@click.option(
    '--rest',
    'rest',
    help='output the rest of subset',
    type=str,
)
@click.option(
    '--base',
    'base_path',
    help='(Advanced) base directory to make test names portable',
    type=click.Path(exists=True, file_okay=False),
    metavar="DIR",
)
@click.option(
    "--same-bin",
    'same_bin_files',
    help="(Advanced) gather specified tests into same bin",
    type=click.Path(),
    multiple=True,
)
@click.pass_context
def split_subset(
        context: click.core.Context,
        subset_id: str,
        bin_target: FractionType,
        rest: str,
        base_path: str,
        same_bin_files: Optional[List[str]],
):
    TestPathWriter.base_path = base_path

    class SplitSubset(TestPathWriter):
        def __init__(self, dry_run: bool = False):
            super(SplitSubset, self).__init__(dry_run=dry_run)

        def run(self):
            index = bin_target[0]
            count = bin_target[1]

            if (index == 0 or count == 0):
                click.echo(
                    click.style(
                        'Error: invalid bin value. Make sure to set over 0 like `--bin 1/2` but set `--bin {}`'.format(
                            bin_target),
                        'yellow'),
                    err=True,
                )
                return

            if count < index:
                click.echo(
                    click.style(
                        'Error: invalid bin value. Make sure to set below 1 like `--bin 1/2`, `--bin 2/2` '
                        'but set `--bin {}`'.format(bin_target),
                        'yellow'),
                    err=True,
                )
                return

            output = []
            rests = []
            is_observation = False

            try:
                client = LaunchableClient(test_runner=context.invoked_subcommand, dry_run=context.obj.dry_run)

                payload = {
                    "sliceCount": count,
                    "sliceIndex": index,
                    "sameBins": [],
                }

                tests_in_files = []

                if same_bin_files is not None and len(same_bin_files) > 0:
                    if self.same_bin_formatter is None:
                        raise ValueError("--same-bin option is supported only for gradle test and go-test. "
                                         "Please remove --same-bin option for the other test runner.")
                    same_bins = []
                    for same_bin_file in same_bin_files:
                        with open(same_bin_file, "r") as f:
                            """
                            A same_bin_file expects to have a list of tests with one test per line.
                            Each line of test gets formatted and packed to sameBins list in payload.
                            E.g.
                                For gradle:
                                ```
                                $ cat same_bin_file.txt
                                example.AddTest
                                example.DivTest
                                example.SubTest
                                ```
                                Formatted:
                                ```
                                "sameBins" [
                                    [
                                        [{"type": "class", "name": "example.AddTest"}],
                                        [{"type": "class", "name": "example.DivTest"}],
                                        [{"type": "class", "name": "example.SubTest"}]
                                    ]
                                ]
                                ```
                            E.g.
                                For gotest:
                                ```
                                $ cat same_bin_file.txt
                                example.BenchmarkGreeting
                                example.ExampleGreeting
                                ```
                                Formatted:
                                ```
                                "sameBins" [
                                    [
                                        [
                                            {"type": "class", "name": "example"},
                                            {"type": "testcase", "name": "BenchmarkGreeting"}
                                        ],
                                        [
                                            {"type": "class", "name": "example"},
                                            {"type": "testcase", "name": "ExampleGreeting"}
                                        ]
                                    ]
                                ]
                                ```
                            """
                            tests = f.readlines()
                            # make a list to set to remove duplicate.
                            tests = list(set([s.strip() for s in tests]))
                            for tests_in_file in tests_in_files:
                                for test in tests:
                                    if test in tests_in_file:
                                        raise ValueError(
                                            "Error: you cannot have one test, {}, in multiple same-bins.".format(test))
                            tests_in_files.append(tests)
                            test_data = [self.same_bin_formatter(s) for s in tests]
                            same_bins.append(test_data)

                    payload["sameBins"] = same_bins

                res = client.request("POST", "{}/slice".format(subset_id), payload=payload)
                res.raise_for_status()

                output = res.json()["testPaths"]
                rests = res.json()["rest"]
                is_observation = res.json().get("isObservation", False)

            except Exception as e:
                if os.getenv(REPORT_ERROR_KEY):
                    raise e
                else:
                    click.echo(e, err=True)
                    click.echo(click.style(
                        "Warning: the service failed to split subset. Falling back to running all tests", fg='yellow'),
                        err=True)
                    return

            if len(output) == 0:
                click.echo(click.style("Error: no tests found in this subset id.", 'yellow'), err=True)
                return

            if is_observation:
                output = output + rests
            if rest:
                if len(rests) == 0:
                    rests.append(output[0])

                self.write_file(rest, rests)

            self.print(output)

    context.obj = SplitSubset(dry_run=context.obj.dry_run)



from unittest import TestCase

from launchable.utils.file_name_pattern import jvm_test_pattern


class FileNameHeuristicTest(TestCase):
    def test_jvm_file_name(self):
        file_names = [
            'LaunchableTest.java',
            'LaunchableTests.java',
            'LaunchableTestCase.java'
        ]
        for file_name in file_names:
            self.assertTrue(jvm_test_pattern.match(file_name))

        file_names = [
            'LaunchableTest2.java',
            'LaunchableTests2.java',
            'LaunchableTestCase2.java'
            'LaunchableTestUtil.java',
            'LaunchableTestss.java',
            'LaunchableTest2Case.java'
        ]
        for file_name in file_names:
            self.assertFalse(jvm_test_pattern.match(file_name))



import gzip
import json
import os
from pathlib import Path
from unittest import mock

import responses  # type: ignore

from tests.cli_test_case import CliTestCase


class BehaveTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/behave/').resolve()

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset(self):
        pipe = "tutorial.feature"
        result = self.cli('subset', '--target', '10%', '--session', self.session, 'behave', input=pipe)
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[0].request.body).decode())

        expected = self.load_json_from_file(self.test_files_dir.joinpath('subset_result.json'))

        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_test(self):
        result = self.cli('record', 'tests', '--session', self.session, 'behave',
                          str(self.test_files_dir) + "/reports/report.xml")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())

        for e in payload["events"]:
            del e["created_at"]

        expected = self.load_json_from_file(self.test_files_dir.joinpath("record_test_result.json"))

        self.assert_json_orderless_equal(expected, payload)



import glob
import os
import sys
import types

import click

from launchable.commands.record.tests import tests as record_tests_cmd
from launchable.commands.split_subset import split_subset as split_subset_cmd
from launchable.commands.subset import subset as subset_cmd


def cmdname(m):
    """figure out the sub-command name from a test runner function"""

    # a.b.cde -> cde
    # xyz -> xyz
    #
    # In python module name the conventional separator is '_' but in command name,
    # it is '-', so we do replace that
    return m[m.rfind('.') + 1:].replace('_', '-')


def wrap(f, group, name=None):
    """
    Wraps a 'plugin' function into a click command and registers it to the given group.

    a plugin function receives the scanner object in its first argument
    """
    if not name:
        name = cmdname(f.__module__)
    d = click.command(name=name)
    cmd = d(click.pass_obj(f))
    group.add_command(cmd)
    return cmd


def subset(f):
    return wrap(f, subset_cmd)


record = types.SimpleNamespace()
record.tests = lambda f: wrap(f, record_tests_cmd)


def split_subset(f):
    return wrap(f, split_subset_cmd)


class CommonSubsetImpls:
    """
    Typical 'subset' implementations that are reusable.
    """

    def __init__(self, module_name):
        self.cmdname = cmdname(module_name)

    def scan_files(self, pattern):
        """
        Suitable for test runners that use files as unit of tests where file names follow a naming pattern.

        :param pattern: file masks that identify test files, such as '*_spec.rb'
        """
        @click.argument('files', required=True, nargs=-1)
        def subset(client, files):
            # client type: Optimize in def lauchable.commands.subset.subset
            def parse(fname: str):
                if os.path.isdir(fname):
                    client.scan(fname, '**/' + pattern)
                elif fname == '@-':
                    # read stdin
                    for l in sys.stdin:
                        parse(l)
                elif fname.startswith('@'):
                    # read response file
                    with open(fname[1:]) as f:
                        for l in f:
                            parse(l)
                else:
                    # assume it's a file
                    client.test_path(fname)

            for f in files:
                parse(f)

            client.run()
        return wrap(subset, subset_cmd, self.cmdname)


class CommonRecordTestImpls:
    """
    Typical 'record tests' implementations that are reusable.
    """

    def __init__(self, module_name):
        self.cmdname = cmdname(module_name)

    def report_files(self, file_mask="*.xml"):
        """
        Suitable for test runners that create a directory full of JUnit report files.

        'record tests' expect JUnit report/XML file names.
        """

        @click.argument('source_roots', required=True, nargs=-1)
        def record_tests(client, source_roots):
            # client type: RecordTests in def launchable.commands.record.tests.tests
            # Accept both file names and GLOB patterns
            # Simple globs like '*.xml' can be dealt with by shell, but
            # not all shells consistently deal with advanced GLOBS like '**'
            # so it's worth supporting it here.
            for root in source_roots:
                match = False
                for t in glob.iglob(root, recursive=True):
                    match = True
                    if os.path.isdir(t):
                        client.scan(t, file_mask)
                    else:
                        client.report(t)

                if not match:
                    # By following the shell convention, if the file doesn't exist or GLOB doesn't match anything,
                    # raise it as an error. Note this can happen for reasons other than a configuration error.
                    # For example, if a build catastrophically failed and no
                    # tests got run.
                    click.echo("No matches found: {}".format(root), err=True)
                    # intentionally exiting with zero
                    return

            client.run()

        return wrap(record_tests, record_tests_cmd, self.cmdname)


class CommonSplitSubsetImpls:
    def __init__(
            self,
            module_name,
            formatter=None,
            seperator=None,
            same_bin_formatter=None,
    ):
        self.cmdname = cmdname(module_name)
        self._formatter = formatter
        self._separator = seperator
        self._same_bin_formatter = same_bin_formatter

    def split_subset(self):
        def split_subset(client):
            # client type: SplitSubset in def
            # lauchable.commands.split_subset.split_subset
            if self._formatter:
                client.formatter = self._formatter

            if self._separator:
                client.separator = self._separator

            if self._same_bin_formatter:
                client.same_bin_formatter = self._same_bin_formatter

            client.run()

        return wrap(split_subset, split_subset_cmd, self.cmdname)



import json
from collections import namedtuple
from typing import Any, Dict, List, TextIO

import dateutil.parser

ChangedFile = namedtuple('ChangedFile', ['path', 'added', 'deleted'])

GitCommit = namedtuple('GitCommit', [
    'commit_hash', 'parents', 'author_email', 'author_time', 'committer_email',
    'committer_time', 'changed_files'
])


def parse_git_log(fp: TextIO) -> List[GitCommit]:
    """Parses the output of a git log command.

    This parses the output of `git log --pretty='format:{"commit": "%H",
    "parents": "%P", "authorEmail": "%ae", "authorTime": "%aI",
    "committerEmail": "%ce", "committerTime": "%cI"}' --numstat`
    """
    ret = []
    meta = {}  # type: Dict[str, Any]
    files = []  # type: List[ChangedFile]
    for idx, line in enumerate(fp):
        line = line.strip()
        if line == '':
            continue
        try:
            if line.startswith('{'):
                if len(meta) != 0:
                    ret.append(GitCommit(changed_files=files, **meta))
                    meta = {}
                    files = []
                d = json.loads(line)
                meta['commit_hash'] = d['commit']
                meta['parents'] = d['parents'].split(' ')
                meta['author_email'] = d['authorEmail']
                meta['author_time'] = dateutil.parser.parse(d['authorTime'])
                meta['committer_email'] = d['committerEmail']
                meta['committer_time'] = dateutil.parser.parse(
                    d['committerTime'])
            elif line.startswith('-'):
                # Ignore binary file changes
                pass
            else:
                added, deleted, path = line.split('\t', 3)
                files.append(ChangedFile(path=path, added=int(added), deleted=int(deleted)))
        except Exception as e:
            raise ValueError("Failed to parse the file at line {}: {}".format(idx + 1, e))
    if len(meta) != 0:
        ret.append(GitCommit(changed_files=files, **meta))
    return ret



import glob
import json
import os
import pathlib
import subprocess
from typing import Generator, List

import click

from launchable.commands.record.case_event import CaseEvent, CaseEventType
from launchable.testpath import TestPath

from . import launchable


# Please specify junit_family=legacy for pytest report format. if using pytest version 6 or higher.
# - pytest has changed its default test report format from xunit1 to xunit2 since version 6.
#     - https://docs.pytest.org/en/latest/deprecations.html#junit-family-default-value-change-to-xunit2
# - The xunit2 format no longer includes file names.
#     - It is possible to output in xunit1 format by specifying junit_family=legacy.
#     - The xunit1 format includes the file name.
# The format of pytest changes depending on the existence of the class. They are also incompatible with
# the junit format.
# Therefore, it converts to junit format at the timing of sending the subset, and converts the returned
# value to pytest format.
# for Example
# $ pytest --collect-only -q
# > tests/test_mod.py::TestClass::test__can_print_aaa
# > tests/fooo/func4_test.py::test_func6
# >
# > 2 tests collected in 0.02s
# result.xml(junit)
# <testcase classname="tests.fooo.func4_test" name="test_func6" file="tests/fooo/func4_test.py"
# line="0" time="0.000" />
# <testcase classname="tests.test_mod.TestClass" name="test__can_print_aaa" file="tests/test_mod.py"
# line="3" time="0.001" />
#
@click.argument('source_roots', required=False, nargs=-1)
@launchable.subset
def subset(client, source_roots: List[str]):
    def _add_testpaths(lines: List[str]):
        for line in lines:
            line = line.rstrip()
            # When an empty line comes, it's done.
            if not line:
                break

            test_path = _parse_pytest_nodeid(line)
            client.test_path(test_path)
    if not source_roots:
        _add_testpaths(client.stdin())
    else:
        command = ["pytest", "--collect-only", "-q"]
        command.extend(source_roots)
        try:
            result = subprocess.run(command, stdout=subprocess.PIPE, universal_newlines=True)
            _add_testpaths(result.stdout.split(os.linesep))
        except FileNotFoundError:
            raise click.ClickException("pytest command not found. Please check the path.")

    client.formatter = _pytest_formatter
    client.run()


def _parse_pytest_nodeid(nodeid: str) -> TestPath:
    data = nodeid.split("::")
    file = data[0]
    class_name = _path_to_class_name(file)
    normalized_file = os.path.normpath(file)

    # file name only
    if len(data) == 1:
        return [
            {"type": "file", "name": normalized_file},
            {"type": "class", "name": class_name},
        ]
    # file + testcase, or file + class + testcase
    else:
        testcase = data[-1]
        if len(data) == 3:
            class_name += "." + data[1]

        return [
            {"type": "file", "name": normalized_file},
            {"type": "class", "name": class_name},
            {"type": "testcase", "name": testcase},
        ]


def _path_to_class_name(path):
    '''
    tests/fooo/func4_test.py -> tests.fooo.func4_test
    '''
    return os.path.splitext(os.path.normpath(path))[0].replace(os.sep, ".")


def _pytest_formatter(test_path):
    for path in test_path:
        t = path.get('type', '')
        n = path.get('name', '')
        if t == 'class':
            cls_name = n
        elif t == 'testcase':
            case = n
        elif t == 'file':
            file = n
    # If there is no class, junitformat use package name, but pytest will be omitted
    # pytest -> tests/fooo/func4_test.py::test_func6
    # junitformat -> <testcase classname="tests.fooo.func4_test"
    # name="test_func6" file="tests/fooo/func4_test.py" line="0" time="0.000"
    # />
    if cls_name == _path_to_class_name(file):
        return "{}::{}".format(file, case)

    else:
        # junitformat's class name includes package, but pytest does not
        # pytest -> tests/test_mod.py::TestClass::test__can_print_aaa
        # junitformat -> <testcase classname="tests.test_mod.TestClass"
        # name="test__can_print_aaa" file="tests/test_mod.py" line="3"
        # time="0.001" />
        return "{}::{}::{}".format(file, cls_name.split(".")[-1], case)


split_subset = launchable.CommonSplitSubsetImpls(__name__, formatter=_pytest_formatter).split_subset()


@click.option('--json', 'json_report', help="use JSON report files produced by pytest-dev/pytest-reportlog",
              is_flag=True)
@click.argument('source_roots', required=True, nargs=-1)
@launchable.record.tests
def record_tests(client, json_report, source_roots):

    ext = "json" if json_report else "xml"
    for root in source_roots:
        match = False
        for t in glob.iglob(root, recursive=True):
            match = True
            if os.path.isdir(t):
                client.scan(t, "*.{}".format(ext))
            else:
                client.report(t)

        if not match:
            click.echo("No matches found: {}".format(root), err=True)
            return

    if json_report:
        client.parse_func = PytestJSONReportParser(client).parse_func

    client.run()


"""
If you want to use --json option, please install pytest-dev/pytest-reportlog.
(https://github.com/pytest-dev/pytest-reportlog)

Usage

```
$ pip install -U pytest-reportlog
$ pytest --report-log report.json
$ launchable record tests --json report.json
```
"""


class PytestJSONReportParser:
    def __init__(self, client):
        self.client = client

    def parse_func(
            self, report_file: str) -> Generator[CaseEventType, None, None]:
        with open(report_file, 'r') as json_file:
            for line in json_file:
                try:
                    data = json.loads(line)
                except Exception as e:
                    raise Exception("Can't read JSON format report file {}. Make sure to confirm report file.".format(
                        report_file)) from e

                nodeid = data.get("nodeid", "")
                if nodeid == "":
                    continue

                when = data.get("when", "")
                outcome = data.get("outcome", "")

                if not (when == "call" or (when == "setup" and outcome == "skipped")):
                    continue

                status = CaseEvent.TEST_FAILED
                if outcome == "passed":
                    status = CaseEvent.TEST_PASSED
                elif outcome == "skipped":
                    status = CaseEvent.TEST_SKIPPED

                test_path = _parse_pytest_nodeid(nodeid)
                for path in test_path:
                    if path.get("type") == "file":
                        path["name"] = pathlib.Path(path["name"]).as_posix()

                yield CaseEvent.create(test_path, data.get("duration", 0), status, None, None, None, None)



import gzip
import json
import os
from pathlib import Path
from unittest import mock

import responses  # type: ignore

from launchable.test_runners import maven
from tests.cli_test_case import CliTestCase


class MavenTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/maven/').resolve()

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset(self):
        result = self.cli('subset', '--target', '10%', '--session',
                          self.session, 'maven', str(self.test_files_dir.joinpath('java/test/src/java/').resolve()))
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[0].request.body).decode())

        expected = self.load_json_from_file(self.test_files_dir.joinpath('subset_result.json'))

        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_from_file(self):
        # if we prepare listed file with slash e.g) com/example/launchable/model/aModelATest.class
        # the test will be failed at Windows environment. So, we generate file
        # path list
        def save_file(list, file_name):
            file = str(self.test_files_dir.joinpath(file_name))
            with open(file, 'w+') as file:
                for l in list:
                    file.write(l.replace(".", os.path.sep) + ".class\n")

        list_1 = ["com.example.launchable.model.a.ModelATest",
                  "com.example.launchable.model.b.ModelBTest",
                  "com.example.launchable.model.b.ModelBTest$SomeInner",
                  "com.example.launchable.model.c.ModelCTest",

                  ]

        list_2 = ["com.example.launchable.service.ServiceATest",
                  "com.example.launchable.service.ServiceATest$Inner1$Inner2",
                  "com.example.launchable.service.ServiceBTest",
                  "com.example.launchable.service.ServiceCTest",
                  ]

        save_file(list_1, "createdFile_1.lst")
        save_file(list_2, "createdFile_2.lst")

        result = self.cli('subset',
                          '--target',
                          '10%',
                          '--session',
                          self.session,
                          'maven',
                          "--test-compile-created-file",
                          str(self.test_files_dir.joinpath("createdFile_1.lst")),
                          "--test-compile-created-file",
                          str(self.test_files_dir.joinpath("createdFile_2.lst")))
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[0].request.body).decode())

        expected = self.load_json_from_file(self.test_files_dir.joinpath('subset_from_file_result.json'))

        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_by_absolute_time(self):
        result = self.cli('subset', '--time', '1h30m', '--session',
                          self.session, 'maven', str(self.test_files_dir.joinpath('java/test/src/java/').resolve()))
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[0].request.body).decode())

        expected = self.load_json_from_file(self.test_files_dir.joinpath('subset_by_absolute_time_result.json'))

        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_by_confidence(self):
        result = self.cli('subset', '--confidence', '90%', '--session',
                          self.session, 'maven', str(self.test_files_dir.joinpath('java/test/src/java/').resolve()))
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[0].request.body).decode())

        expected = self.load_json_from_file(self.test_files_dir.joinpath('subset_by_confidence_result.json'))

        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_test_maven(self):
        result = self.cli('record', 'tests', '--session', self.session,
                          'maven', str(self.test_files_dir) + "/**/reports")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())

        for e in payload["events"]:
            del e["created_at"]

        expected = self.load_json_from_file(self.test_files_dir.joinpath("record_test_result.json"))

        self.assert_json_orderless_equal(expected, payload)

    def test_glob(self):
        for x in [
            'foo/BarTest.java',
            'foo/BarTest.class',
            'FooTest.class',
            'TestFoo.class',
        ]:
            self.assertTrue(maven.is_file(x))

        for x in [
            'foo/Bar$Test.class',
            'foo/MyTest$Inner.class',
            'foo/Util.class',
        ]:
            self.assertFalse(maven.is_file(x))



from os.path import join
from typing import Callable, Dict, List, Optional

import click

from ..testpath import TestPath


class TestPathWriter(object):
    base_path = Optional[str]

    def __init__(self, dry_run=False):
        self._formatter = TestPathWriter.default_formatter
        self._separator = "\n"
        self._same_bin_formatter = None
        self.dry_run = dry_run

    @classmethod
    def default_formatter(cls, x: TestPath):
        """default formatter that's in line with to_test_path(str)"""
        file_name = x[0]['name']
        if cls.base_path:
            # default behavior consistent with default_path_builder's relative
            # path handling
            file_name = join(str(cls.base_path), file_name)
        return file_name

    @property
    def formatter(self) -> Callable[[TestPath], str]:
        """
        This function, if supplied, is used to format test names
        from the format Launchable uses to the format test runners expect.
        """
        return self._formatter

    @formatter.setter
    def formatter(self, v: Callable[[TestPath], str]):
        self._formatter = v

    @property
    def separator(self) -> str:
        return self._separator

    @separator.setter
    def separator(self, s: str):
        self._separator = s

    def write_file(self, file: str, test_paths: List[TestPath]):
        open(file, "w+", encoding="utf-8").write(
            self.separator.join(self.formatter(t) for t in test_paths))

    def print(self, test_paths: List[TestPath]):
        click.echo(self.separator.join(self.formatter(t)
                                       for t in test_paths))

    @property
    def same_bin_formatter(self) -> Callable[[str], Dict[str, str]]:
        return self._same_bin_formatter

    @same_bin_formatter.setter
    def same_bin_formatter(self, v: Callable[[str], Dict[str, str]]):
        self._same_bin_formatter = v



import os
import tempfile
from unittest import mock

import responses  # type: ignore

from launchable.utils.http_client import get_base_url
from tests.cli_test_case import CliTestCase


class SplitSubsetTest(CliTestCase):
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_split_subset_with_observation_mode(self):
        pipe = "test_1.py\ntest_2.py\ntest_3.py\ntest_4.py\ntest_5.py\ntest_6.py"
        mock_json_response = {
            "testPaths": [
                [{"type": "file", "name": "test_1.py"}],
                [{"type": "file", "name": "test_3.py"}],

            ],
            "rest": [
                [{"type": "file", "name": "test_5.py"}],

            ],
            "subsettingId": 456,
            "isObservation": False,
        }

        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset/{}/slice".format(
                get_base_url(),
                self.organization,
                self.workspace,
                self.subsetting_id),
            json=mock_json_response,
            status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli("split-subset", "--subset-id", "subset/456", "--bin", "1/2", "--rest",
                          rest.name, "file", mix_stderr=False, input=pipe)
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, "test_1.py\ntest_3.py\n")
        self.assertEqual(rest.read().decode(), os.linesep.join(["test_5.py"]))
        rest.close()
        os.unlink(rest.name)

        mock_json_response["isObservation"] = True
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset/{}/slice".format(
                get_base_url(),
                self.organization,
                self.workspace,
                self.subsetting_id),
            json=mock_json_response,
            status=200)

        observation_mode_rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli("split-subset", "--subset-id", "subset/456", "--bin", "1/2", "--rest",
                          observation_mode_rest.name, "file", mix_stderr=False, input=pipe)

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, "test_1.py\ntest_3.py\ntest_5.py\n")
        self.assertEqual(observation_mode_rest.read().decode(), os.linesep.join(["test_5.py"]))
        observation_mode_rest.close()
        os.unlink(observation_mode_rest.name)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_split_subset(self):
        pipe = "test_1.py\ntest_2.py\ntest_3.py\ntest_4.py\ntest_5.py\ntest_6.py"
        mock_json_response = {
            "testPaths": [
                [{"type": "file", "name": "test_1.py"}],
                [{"type": "file", "name": "test_6.py"}],
            ],
            "rest": [],
            "subsettingId": 456,
            "isObservation": False,
        }

        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset/{}/slice".format(
                get_base_url(),
                self.organization,
                self.workspace,
                self.subsetting_id),
            json=mock_json_response,
            status=200)

        result = self.cli(
            "split-subset",
            "--subset-id",
            "subset/456",
            "--bin",
            "1/2",
            "file",
            mix_stderr=False,
            input=pipe,
        )
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, "test_1.py\ntest_6.py\n")



from . import launchable

subset = launchable.CommonSubsetImpls(__name__).scan_files('*_spec.rb')
split_subset = launchable.CommonSplitSubsetImpls(__name__).split_subset()
record_tests = launchable.CommonRecordTestImpls(__name__).report_files()



import os
import sys
from http import HTTPStatus
from typing import Dict, List

import click
from tabulate import tabulate

from ...utils.env_keys import REPORT_ERROR_KEY
from ...utils.http_client import LaunchableClient


@click.command()
@click.option(
    '--subset-id',
    'subset_id',
    help='subest id',
    required=True,
)
def subset(subset_id: int):
    subset = []
    rest = []
    try:
        client = LaunchableClient()
        res = client.request("get", "subset/{}".format(subset_id))

        if res.status_code == HTTPStatus.NOT_FOUND:
            click.echo(click.style(
                "Subset {} not found. Check subset ID and try again.".format(subset_id), 'yellow'), err=True)
            sys.exit(1)

        res.raise_for_status()
        subset = res.json()["testPaths"]
        rest = res.json()["rest"]
    except Exception as e:
        if os.getenv(REPORT_ERROR_KEY):
            raise e
        else:
            click.echo(e, err=True)
        click.echo(click.style(
            "Warning: failed to inspect subset", fg='yellow'),
            err=True)

    header = ["Order", "Test Path", "In Subset", "Estimated duration (sec)"]

    subset_row = convert_row(subset, 1, True)
    rest_row = convert_row(rest, len(subset) + 1, False)
    rows = subset_row + rest_row

    click.echo(tabulate(rows, header, tablefmt="github"))


def convert_row(data_list: List[Dict], order: int, is_subset: bool):
    """
    data_list: testPaths or rest in response to a get subset API
    order: start number of order
    is_subset: in subset or not
    """
    data = []
    for i, l in enumerate(data_list):
        if l.keys() >= {"testPath"}:
            data.append(
                [
                    order + i,
                    "#".join([path["type"] + "=" + path["name"]
                              for path in l["testPath"] if path.keys() >= {"type", "name"}]),
                    "✔" if is_subset else "",
                    "{:0.4f}".format(l.get("duration", 0.0) / 1000),
                ]
            )
    return data



import gzip
import json
import os
import shutil
import tempfile
import types
import unittest

import click  # type: ignore
import responses  # type: ignore
from click.testing import CliRunner  # type: ignore

from launchable.__main__ import main
from launchable.utils.http_client import get_base_url
from launchable.utils.session import SESSION_DIR_KEY, clean_session_files


class CliTestCase(unittest.TestCase):
    """
    Base class for setting up test to invoke CLI
    """
    organization = 'launchableinc'
    workspace = 'mothership'
    launchable_token = "v1:{}/{}:auth-token-sample".format(organization, workspace)
    session_id = 16
    build_name = "123"
    subsetting_id = 456
    session = "builds/{}/test_sessions/{}".format(build_name, session_id)

    def setUp(self):
        self.dir = tempfile.mkdtemp()
        os.environ[SESSION_DIR_KEY] = self.dir

        self.maxDiff = None

        responses.add(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/builds/{}/test_sessions".format(
                get_base_url(),
                self.organization,
                self.workspace,
                self.build_name),
            json={'id': self.session_id},
            status=200)
        responses.add(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(),
                self.organization,
                self.workspace),
            json={'testPaths': [], 'rest': [], 'subsettingId': 456},
            status=200)
        responses.add(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset/{}/slice".format(
                get_base_url(),
                self.organization,
                self.workspace,
                self.subsetting_id),
            json={'testPaths': [], 'rest': [], 'subsettingId': 456},
            status=200)
        responses.add(
            responses.GET,
            "{}/intake/organizations/{}/workspaces/{}/subset/{}".format(
                get_base_url(),
                self.organization,
                self.workspace,
                self.subsetting_id),
            json={'testPaths': [], 'rest': [], 'subsettingId': 456},
            status=200)
        responses.add(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/builds/{}/test_sessions/{}/events".format(
                get_base_url(),
                self.organization,
                self.workspace,
                self.build_name,
                self.session_id),
            json={},
            status=200)
        responses.add(
            responses.GET,
            "{}/intake/organizations/{}/workspaces/{}/test_sessions/{}/events".format(
                get_base_url(),
                self.organization,
                self.workspace,
                self.session_id),
            json=[],
            status=200)
        responses.add(
            responses.GET,
            "{}/intake/organizations/{}/workspaces/{}/builds/{}/test_sessions/{}".format(
                get_base_url(),
                self.organization,
                self.workspace,
                self.build_name,
                self.session_id),
            json={
                'id': self.session_id,
                'isObservation': False,
            },
            status=200)
        responses.add(
            responses.PATCH,
            "{}/intake/organizations/{}/workspaces/{}/builds/{}/test_sessions/{}/close".format(
                get_base_url(),
                self.organization,
                self.workspace,
                self.build_name,
                self.session_id),
            json={},
            status=200)
        responses.add(
            responses.GET,
            "{}/intake/organizations/{}/workspaces/{}/builds/{}".format(
                get_base_url(),
                self.organization,
                self.workspace,
                self.build_name),
            json={'createdAt': "2020-01-02T03:45:56.123+00:00", 'id': 123},
            status=200)
        responses.add(
            responses.GET,
            "{}/intake/organizations/{}/workspaces/{}/slack/notification/key/list".format(
                get_base_url(),
                self.organization,
                self.workspace),
            json={'keys': ["GITHUB_ACTOR", "BRANCH_NAME"]},
            status=200)

    def tearDown(self):
        clean_session_files()
        del os.environ[SESSION_DIR_KEY]
        shutil.rmtree(self.dir)

    def cli(self, *args, **kwargs) -> click.testing.Result:
        # for CliRunner kwargs
        mix_stderr = True
        if 'mix_stderr' in kwargs:
            mix_stderr = kwargs['mix_stderr']
            del kwargs['mix_stderr']

        """
        Invoke CLI command and returns its result
        """
        return CliRunner(mix_stderr=mix_stderr).invoke(main, args, catch_exceptions=False, **kwargs)

    def load_json_from_file(self, file):
        try:
            with file.open() as json_file:
                return json.load(json_file)
        except Exception as e:
            raise IOError("Failed to parse JSON {}".format(file)) from e

    def payload(self, mock_post):
        """
        Given a mock request object, capture the payload sent to the server
        """
        for (args, kwargs) in mock_post.call_args_list:
            if kwargs.get('data'):
                data = kwargs['data']
                if isinstance(data, types.GeneratorType):
                    data = b''.join(data)
                return data

        raise Exception("No data posted")

    def gzipped_json_payload(self, mock_post):
        return json.loads(gzip.decompress(self.payload(mock_post)).decode())

    def json_payload(self, mock_post):
        return json.loads(self.payload(mock_post))

    def assert_json_orderless_equal(self, a, b):
        """
        Compare two JSON trees ignoring orders of items in list & dict
        """
        def tree_sorted(obj):
            if isinstance(obj, dict):
                return sorted((k, tree_sorted(v)) for k, v in obj.items())
            if isinstance(obj, list):
                return sorted(tree_sorted(x) for x in obj)
            else:
                return obj

        self.assertEqual(tree_sorted(a), tree_sorted(b))



import json
import os
import threading
from http.server import HTTPServer, SimpleHTTPRequestHandler
from unittest import mock

from launchable.commands.record.commit import _build_proxy_option
from launchable.utils.env_keys import BASE_URL_KEY
from tests.cli_test_case import CliTestCase


class CommitHandler(SimpleHTTPRequestHandler):
    # mock commits/latest
    def do_GET(self):
        response = []
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps(response).encode("utf-8"))

    # mock commits/collect
    def do_POST(self):
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.end_headers()


class CommitTest(CliTestCase):

    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_run_commit(self):
        server = HTTPServer(("", 0), CommitHandler)
        thread = threading.Thread(None, server.serve_forever)
        thread.start()

        host, port = server.server_address
        endpoint = "http://{}:{}".format(host, port)

        with mock.patch.dict(os.environ, {BASE_URL_KEY: endpoint}):
            result = self.cli("record", "commit")
            self.assertEqual(result.exit_code, 0, "exit normally")

        server.shutdown()
        thread.join()

    def test_proxy_options(self):
        self.assertEqual(_build_proxy_option("https://some_proxy:1234"),
                         ['-Dhttps.proxyHost=some_proxy', '-Dhttps.proxyPort=1234'])
        self.assertEqual(_build_proxy_option("some_proxy:1234"), ['-Dhttps.proxyHost=some_proxy', '-Dhttps.proxyPort=1234'])
        self.assertEqual(_build_proxy_option("some_proxy"), ['-Dhttps.proxyHost=some_proxy'])
        self.assertEqual(_build_proxy_option("https://some_proxy"), ['-Dhttps.proxyHost=some_proxy'])
        self.assertEqual(_build_proxy_option("http://yoyoyo"), ['-Dhttps.proxyHost=yoyoyo'])



import os
import shutil
import subprocess
import sys


def get_java_command():
    if shutil.which("java"):
        return "java"

    if os.access(os.path.expandvars("$JAVA_HOME/bin/java"), os.X_OK):
        return os.path.expandvars("$JAVA_HOME/bin/java")

    return None


def cygpath(p):
    # When running in Cygwin ported Python (as opposed to Windows native Python), the paths we deal with are in
    # the cygwin format. But when we invoke Windows native Java (and there's no Cygwin ported Java), those parameters
    # need to be in the Windows path format.
    #
    # In Cygwin aware world, it is always the responsibility of the Cygwin process calling Windows native process to
    # do this conversion, so here we are.
    #
    # Cygwin ported Python is not to be confused with Windows native Python running in Windows with cygwin.
    # So tests like CYGWIN env var, "uname" are incorrect.
    if sys.platform == 'cygwin':
        p = subprocess.check_output(['cygpath', '-w', p]).decode().strip()
    return p




import os
from io import StringIO
from unittest import mock

import responses  # type: ignore

from launchable.commands.helper import _check_observation_mode_status
from launchable.utils.http_client import get_base_url
from tests.cli_test_case import CliTestCase


class HelperTest(CliTestCase):

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_check_observation_mode_status(self):
        test_session = "builds/{}/test_sessions/{}".format(
            self.build_name,
            self.session_id,
        )

        with mock.patch('sys.stderr', new=StringIO()) as stderr:
            _check_observation_mode_status(test_session, False)
            print(stderr.getvalue())
            self.assertNotIn("WARNING:", stderr.getvalue())

        request_path = "{}/intake/organizations/{}/workspaces/{}/{}".format(
            get_base_url(),
            self.organization,
            self.workspace,
            test_session,
        )

        with mock.patch('sys.stderr', new=StringIO()) as stderr:
            responses.replace(
                responses.GET,
                request_path,
                json={
                    "isObservation": False
                }, status=200)

            _check_observation_mode_status(test_session, True)
            self.assertIn("WARNING:", stderr.getvalue())

        with mock.patch('sys.stderr', new=StringIO()) as stderr:
            responses.replace(
                responses.GET,
                request_path,
                json={
                    "isObservation": True
                }, status=200)

            _check_observation_mode_status(test_session, True)
            self.assertNotIn("WARNING:", stderr.getvalue())

        with mock.patch('sys.stderr', new=StringIO()) as stderr:
            responses.replace(
                responses.GET,
                request_path,
                json={
                    "isObservation": True
                }, status=404)

            _check_observation_mode_status(test_session, False)

            # not check when status isn't 200
            self.assertNotIn("WARNING:", stderr.getvalue())
